// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

#ifndef _ERRORSTRINGS_H_
#define _ERRORSTRINGS_H_

LPCWSTR GetPalErrorString(DWORD code);

#endif // _ERRORSTRINGS_H_
