// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.


//This file just includes formattype.cpp from the inc directory.
#include "common.h"

#include "../inc/formattype.cpp"
