// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.


//


#include "stdafx.h"

#include "../../shared/amd64/primitives.cpp"


