// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.
//
// File: external.h
//

//
// External types used in MetaData\* subcomponent classes.
//
// ======================================================================================

#pragma once

#include <clrtypes.h>
#include <safemath.h>
#include <new>

using std::nothrow;
