// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

#include "common.h"
#include "yieldprocessornormalized.h"


#include "finalizerthread.h"

#include "yieldprocessornormalizedshared.cpp"
