// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.
#include "common.h"
#include "asmconstants.h"

#include "../i386/stublinkerx86.cpp"

