// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

//

/*
 *
 * Common source file for all files in ..\shared for compiling into the right-side
 *
 */
#include "stdafx.h"

#include "../shared/utils.cpp"
#include "../shared/dbgtransportsession.cpp"
#include "../shared/stringcopyholder.cpp"
