// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.


#ifndef __CGROUP_H__
#define __CGROUP_H__

void InitializeCGroup();
void CleanupCGroup();

#endif // __CGROUP_H__

