// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.


//
 /*vim: set foldmethod=marker: */
#include <stdafx.h>

HRESULT ClrDataAccess::DumpNativeImage(CLRDATA_ADDRESS loadedBase,
    LPCWSTR name,
    IXCLRDataDisplay* display,
    IXCLRLibrarySupport* support,
    IXCLRDisassemblySupport* dis)
{
    return E_FAIL;
}
