// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

#pragma once
#include "pal_types.h"
#include "opensslshim.h"

typedef STACK_OF(X509) X509Stack;
typedef STACK_OF(X509_NAME) X509NameStack;
