// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

//
// This is file to compile 2 public "headers" from ..\inc directory (we have to include stdafx.h first).
//
/***************************************************************************/

#include "stdafx.h"

#include "../inc/corhlpr.cpp"
#include "../inc/corhlprpriv.cpp"
