// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

BSOPNAME(BSOP_Assign)
BSOPNAME(BSOP_AssignAllowUninitRhs)
BSOPNAME(BSOP_AssignNocopy)
BSOPNAME(BSOP_OldStyleClearD)
BSOPNAME(BSOP_ClearD)
BSOPNAME(BSOP_MakeSingleton)
BSOPNAME(BSOP_MakeEmpty)
BSOPNAME(BSOP_MakeFull)
BSOPNAME(BSOP_MakeCopy)
BSOPNAME(BSOP_IsEmpty)
BSOPNAME(BSOP_Count)
BSOPNAME(BSOP_RemoveElemD)
BSOPNAME(BSOP_RemoveElem)
BSOPNAME(BSOP_AddElemD)
BSOPNAME(BSOP_AddElem)
BSOPNAME(BSOP_UnionD)
BSOPNAME(BSOP_UnionDChanged)
BSOPNAME(BSOP_Union)
BSOPNAME(BSOP_IntersectionD)
BSOPNAME(BSOP_Intersection)
BSOPNAME(BSOP_IsEmptyIntersection)
BSOPNAME(BSOP_DiffD)
BSOPNAME(BSOP_Diff)
BSOPNAME(BSOP_IsMember)
BSOPNAME(BSOP_IsNotMember)
BSOPNAME(BSOP_NoBitsAbove)
BSOPNAME(BSOP_LeftShiftSingletonByOneD)
BSOPNAME(BSOP_IsSubset)
BSOPNAME(BSOP_Equal)
BSOPNAME(BSOP_NotEqual)
BSOPNAME(BSOP_NextBit)
BSOPNAME(BSOP_ToString)
