// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

#ifndef __GLOBALS_H__
#define __GLOBALS_H__

const int tccSecondsToMilliSeconds = 1000;

// The number of microseconds in a second.
const int tccSecondsToMicroSeconds = 1000000;

// The number of nanoseconds in a second.
const int tccSecondsToNanoSeconds = 1000000000;

// The number of microseconds in a millisecond.
const int tccMilliSecondsToMicroSeconds = 1000;

// The number of nanoseconds in a millisecond.
const int tccMilliSecondsToNanoSeconds = 1000000;

#endif // __GLOBALS_H__
